/*
* Copyright 2009, KyTea Development Team
* 
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
* 
*     http://www.apache.org/licenses/LICENSE-2.0
* 
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

#ifndef KYTEA_MODEL_H__
#define KYTEA_MODEL_H__

#include "liblinear/linear.h"
#include "kytea-string.h"
#include <tr1/unordered_map>
#include <map>
#include <cmath>

#define KYTEA_SAFE
#define SIG_CUTOFF 1E-6

namespace kytea {

class KyteaModel {
public:

    typedef std::tr1::unordered_map<KyteaString,unsigned,KyteaStringHash> FeatMap;
    typedef vector<KyteaString> FeatVec;

private:

    FeatMap ids_;
    FeatVec names_;
    model * mod_;
    bool addFeat_;

public:
    KyteaModel() : mod_(0), addFeat_(true) {
        KyteaString str;
        mapFeat(str);
    }
    ~KyteaModel() {
        if(mod_)
            destroy_model(mod_);
    }

    // feature functions
    inline unsigned mapFeat(const KyteaString & str) {
        FeatMap::const_iterator it = ids_.find(str);
        unsigned ret = 0;
        if(it != ids_.end())
            ret = it->second;
        else if(addFeat_) {
            ret = names_.size();
            ids_[str] = ret;
            names_.push_back(str);
        }
        return ret;
    }
    KyteaString showFeat(unsigned val) {
#ifdef MODEL_SAFE
        if(val >= names_.size())
            throw runtime_error("FATAL: Array index out of bounds in showFeat");
#endif
        return names_[val];
    }

    int getBiasId() {
        return names_.size();
    }

    void setModel(model * mod) { mod_ = mod; }
    void setAddFeatures(bool addFeat) { addFeat_ = addFeat; }

    const FeatVec & getNames() const { return names_; }
    const model * getModel() const { return mod_; }

    pair<int,double> runClassifier(const vector<unsigned> & feat) {
        if(!mod_)
            throw runtime_error("FATAL: Attempt to run classifier when no model exists");
        bool hasBias = (mod_->bias >= 0);
        feature_node nodes[feat.size()+(hasBias?2:1)];
        unsigned j;
        for(j = 0; j < feat.size(); j++) {
            nodes[j].index = feat[j];
            nodes[j].value = 1;
        }
        if(hasBias) {
            nodes[j].index = names_.size();
            nodes[j++].value = mod_->bias;
        }
        nodes[j].index = -1;
        // make the prediction
        double *dec_values = (double*)malloc(sizeof(double)*mod_->nr_class);
        // double *dec_values = malloc(double, sizeof(double)*model_->nr_class);
        pair<int,double> ret;
        ret.first=predict_values(mod_, nodes, dec_values);
        if(mod_->nr_class == 2)
            ret.second = abs(dec_values[0]);
        else {
            double conf = 100;
            for(j = 0; j < (unsigned)mod_->nr_class && ret.first != mod_->label[j]; j++);
            for(unsigned k = 0; k < (unsigned)mod_->nr_class; k++)
                if(k != j)
                    ret.second = min(conf,dec_values[j]-dec_values[k]);
        }
        free(dec_values); 
        return ret;
    }

    void trainModel(vector<feature_node*> & xs, vector<int> & ys, double bias, int solver, double epsilon) {
        if(mod_) 
            delete mod_;
        mod_ = 0;
        // build the liblinear model
        struct problem   prob;
        struct parameter param;
        prob.l = xs.size();
        prob.y = &ys.front();
        prob.x = &xs.front();
        prob.bias = bias;
        prob.n = getBiasId()+(prob.bias>=0?1:0);
        if(prob.bias >= 0) {
            int biasId = getBiasId();
            for(int i = 0; i < prob.l; i++) {
                feature_node * ptr = prob.x[i];
                while(ptr->index != -1)
                    ptr++;
                ptr->index = biasId;
            }
        }

        param.solver_type = solver;
        param.C = 1; // L2R_L2LOSS_SVC_DUAL;
        param.eps = epsilon;
        param.nr_weight = 0;
        param.weight_label = NULL;
        param.weight = NULL;
	    if(param.eps == HUGE_VAL) {
	    	if(param.solver_type == L2R_LR || param.solver_type == L2R_L2LOSS_SVC)
	    		param.eps = 0.01;
	    	else if(param.solver_type == L2R_L2LOSS_SVC_DUAL || param.solver_type == L2R_L1LOSS_SVC_DUAL || param.solver_type == MCSVM_CS)
	    		param.eps = 0.1;
	    	else if(param.solver_type == L1R_L2LOSS_SVC || param.solver_type == L1R_LR)
	    		param.eps = 0.01;
	    }
        mod_ = train(&prob, &param);
        trim();
    }

    void trim(double sig = SIG_CUTOFF) {
        if(!mod_) return;

        int beforeTrim = names_.size()-1, nr_w, i, j, afterTrim = 0;
        if(mod_->nr_class==2 && mod_->param.solver_type != MCSVM_CS)
        	nr_w=1;
        else
        	nr_w=mod_->nr_class;

        // find values to trim
        vector<bool> preserve(beforeTrim);
        for(i=0; i<beforeTrim; i++) {
            double myMax = 0.0;
        	for(j=0; j<nr_w; j++)
                myMax = max(abs(mod_->w[i*nr_w+j]),myMax);
            preserve[i] = (myMax>sig);
            if(preserve[i])
                afterTrim++;
        }

        // make new model
        model * newModel = (model*)malloc(sizeof(model));
        FeatVec newNames(afterTrim+1);
        ids_.clear();
        KyteaString empty;
        newNames[0] = empty;
        ids_[empty] = 0;
        newModel->param.solver_type = mod_->param.solver_type;
        newModel->nr_class = mod_->nr_class;
        newModel->nr_feature = afterTrim;
        int w_size = (mod_->bias>=0?afterTrim+1:afterTrim);
        newModel->w = (double*)malloc(nr_w*w_size*sizeof(double));
        int curr=0, entrySize=sizeof(double)*nr_w;
        for(i=0; i<beforeTrim; i++) {
            if(preserve[i]) {
                newNames[curr+1] = names_[i+1];
                ids_.insert(pair<KyteaString,unsigned>(names_[i],(unsigned)curr));
                memcpy(newModel->w+(curr++)*nr_w, mod_->w+i*nr_w, entrySize);
            }
        }
        // get the bias
        if(curr < w_size)
            memcpy(newModel->w+curr*nr_w, mod_->w+i*nr_w, entrySize);
        newModel->label = (int*)malloc(mod_->nr_class*sizeof(int));
        memcpy(newModel->label, mod_->label, mod_->nr_class*sizeof(int));
        newModel->bias = mod_->bias;

        names_=newNames;
        destroy_model(mod_);
        mod_ = newModel;

    }

};

}

#endif
