/*
* Copyright 2009, KyTea Development Team
* 
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
* 
*     http://www.apache.org/licenses/LICENSE-2.0
* 
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

#include "model-io.h"
#include <algorithm>

#define BUFFER_SIZE 4096

using namespace kytea;

static const char *solver_type_table[]=
{
	"L2R_LR", "L2R_L2LOSS_SVC_DUAL", "L2R_L2LOSS_SVC","L2R_L1LOSS_SVC_DUAL","MCSVM_CS", "L1R_L2LOSS_SVC","L1R_LR", NULL
};

ModelIO * ModelIO::createIO(const char* file, Format form, bool output, KyteaConfig & config) {
    if(output && form == ModelIO::FORMAT_UNKNOWN)
        throw runtime_error("A format must be specified for model output");
    else if(!output) {
        ifstream ifs(file);
        if(!ifs.good()) 
            throw runtime_error("Could not open model file");
        string line,buff1,buff2,buff3,buff4;
        getline(ifs, line);
        istringstream iss(line);
        if(!(iss >> buff1) || !(iss >> buff2) || !(iss >> buff3) || !(iss >> buff4) || 
                                  buff1 != "KyTea" || buff3.length() != 1)
            throw runtime_error("Badly formed model (header incorrect)");
        if(buff2 != MODEL_IO_VERSION)
            throw runtime_error("Incompatible model version");
        form = buff3[0];
        config.setEncoding(buff4.c_str());
        ifs.close();
    }
    StringUtil * util = config.getStringUtil();
    if(form == ModelIO::FORMAT_TEXT)      { return new TextModelIO(util,file,output); }
    else if(form == ModelIO::FORMAT_BINARY) { return new BinaryModelIO(util,file,output); }
    else {
        throw runtime_error("Illegal model format");
    }
}

ModelIO * ModelIO::createIO(iostream & file, Format form, bool output, KyteaConfig & config) {
    StringUtil * util = config.getStringUtil();
    if(form == ModelIO::FORMAT_TEXT)      { return new TextModelIO(util,file,output); }
    else if(form == ModelIO::FORMAT_BINARY) { return new BinaryModelIO(util,file,output); }
    else {
        throw runtime_error("Illegal model format");
    }
}

void TextModelIO::writeConfig(const KyteaConfig & config) {

    *str_ << "KyTea " << MODEL_IO_VERSION << " T " << config.getEncodingString() << endl;

    if(!config.getDoWS()) *str_ << "-nows" << endl;
    if(!config.getDoPE()) *str_ << "-nope" << endl;
    *str_ << "-charw " << (int)config.getCharWindow() << endl;
    *str_ << "-charn " << (int)config.getCharN() << endl;
    *str_ << "-typew " << (int)config.getTypeWindow() << endl;
    *str_ << "-typen " << (int)config.getTypeN() << endl;
    *str_ << "-dicn "  << (int)config.getDictionaryN() << endl;
    *str_ << "-bias " << config.getBias() << endl;
    *str_ << "-eps " << config.getEpsilon() << endl;
    *str_ << "-solver " << config.getSolverType() << endl;

    *str_ << endl;

}

void TextModelIO::readConfig(KyteaConfig & config) {
    string line,s1,s2;
    getline(*str_,line); // ignore the header
    while(getline(*str_, line) && line.length() != 0) {
        istringstream iss(line);
        iss >> s1;
        iss >> s2;
        config.parseTrainArg(s1.c_str(), (s2.length()==0?0:s2.c_str()));
    }
}

void TextModelIO::writeModel(const KyteaModel * mod) {

    // print a single endl for empty models
    if(mod == 0) {
        *str_ << endl;
        return;
    }

    // print the feature names
    const KyteaModel::FeatVec & names = mod->getNames();
    *str_ << names.size()-1 << endl;
    for(unsigned i = 1; i < names.size(); i++)
        *str_ << util_->showString(names[i]) << endl;

    const model * model_ = mod->getModel();

    int i;
    int nr_feature=model_->nr_feature;
    int n;
    const parameter& param = model_->param;

    if(model_->bias>=0)
    	n=nr_feature+1;
    else
    	n=nr_feature;
    int w_size = n;

    int nr_w;
    if(model_->nr_class==2 && model_->param.solver_type != MCSVM_CS)
    	nr_w=1;
    else
    	nr_w=model_->nr_class;

    *str_ << "solver_type " << solver_type_table[param.solver_type] << endl;
    *str_ << "nr_class " << model_->nr_class << endl;
    *str_ << "label";
    for(i=0; i<model_->nr_class; i++)
        *str_ << " " << model_->label[i];    
    *str_ << endl;

    *str_ << "nr_feature " << nr_feature << endl;

    char buffer[50];

    // fix this
    sprintf(buffer, "%.16g", model_->bias);
    *str_ << "bias " << buffer << endl;

    *str_ << "w" << endl;
    for(i=0; i<w_size; i++)
    {
    	int j;
    	for(j=0; j<nr_w; j++) {
            sprintf(buffer, "%.16g", model_->w[i*nr_w+j]);
            *str_ << buffer << " ";
        }
        *str_ << endl;
    }

    *str_ << endl;
}

KyteaModel * TextModelIO::readModel() {

    // the first line either contains the feature count or empty line
    string line;
    getline(*str_, line);
    if(line.length() == 0)
        return 0;

	int i;
	int nr_feature;
	int n;
	int nr_class;
	double bias;
	model *model_ = (model*)malloc(sizeof(model));
    KyteaModel * mod = new KyteaModel();
    mod->setModel(model_);
	parameter& param = model_->param;

    i = util_->parseInt(line.c_str());
    for( ; i > 0; i--) {
        getline(*str_, line);
        mod->mapFeat(util_->mapString(line));
    }

	model_->label = NULL;

	string str;
	while(1)
	{
        *str_ >> str;
		if(strcmp(str.c_str(),"solver_type")==0)
		{
            *str_ >> str;
			int i;
			for(i=0;solver_type_table[i];i++)
			{
				if(strcmp(solver_type_table[i],str.c_str())==0)
				{
					param.solver_type=i;
					break;
				}
			}
			if(solver_type_table[i] == NULL)
			{
				free(model_->label);
				free(model_);
				throw runtime_error("unknown solver type.");
			}
		}
		else if(strcmp(str.c_str(),"nr_class")==0)
		{
            *str_ >> nr_class;
			model_->nr_class=nr_class;
		}
		else if(strcmp(str.c_str(),"nr_feature")==0)
		{
            *str_ >> nr_feature;
			model_->nr_feature=nr_feature;
		}
		else if(strcmp(str.c_str(),"bias")==0)
		{
            *str_ >> bias;
			model_->bias=bias;
		}
		else if(strcmp(str.c_str(),"w")==0)
		{
            // clear out the rest of the line
            getline(*str_,str);
			break;
		}
		else if(strcmp(str.c_str(),"label")==0)
		{
			int nr_class = model_->nr_class;
			model_->label = (int*)malloc(sizeof(int)*nr_class);
			for(int i=0;i<nr_class;i++)
				*str_ >> model_->label[i];
		}
		else
		{
			free(model_);
            ostringstream err;
            err << "Unknown text in model file '" << str << "'";
			throw runtime_error(err.str());
		}
	}

	nr_feature=model_->nr_feature;
	if(model_->bias>=0)
		n=nr_feature+1;
	else
		n=nr_feature;
	int w_size = n;
	int nr_w;
	if(nr_class==2 && param.solver_type != MCSVM_CS)
		nr_w = 1;
	else
		nr_w = nr_class;

	model_->w=(double*)malloc(sizeof(double)*w_size*nr_w);
	for(i=0; i<w_size; i++)
	{
		int j;
        getline(*str_,str);
        istringstream iss(str);
        string buff;
		for(j=0; j<nr_w; j++) {
            iss >> buff;
            model_->w[i*nr_w+j] = util_->parseFloat(buff.c_str());
        }
	}
    
    getline(*str_, str);
    if(str.length() != 0 && str != " ") {
        ostringstream err;
        err << "Bad line when expecting end of file: '" << str << "'" << endl;
        throw runtime_error(err.str());
    }

    // read models shouldn't add any additional features
    mod->setAddFeatures(false);

    return mod;

}

void TextModelIO::writeDictionary(const Dictionary * dict) {
    // write the states
    const vector<DictionaryState*> & states = dict->getStates();
    *str_ << states.size() << endl;
    for(unsigned i = 0; i < states.size(); i++) {
        *str_ << states[i]->failure;
        for(unsigned j = 0; j < states[i]->gotos.size(); j++)
            *str_ << " " << util_->showChar(states[i]->gotos[j].first) << " " << states[i]->gotos[j].second;
        *str_ << endl;
        for(unsigned j = 0; j < states[i]->output.size(); j++) {
            if(j!=0) *str_ << " ";
            *str_ << states[i]->output[j];
        }
        *str_ << endl;
        *str_ << (states[i]->isBranch?'b':'n') << endl;
    }
    // write the entries
    const vector<DictionaryEntry*> & entries = dict->getEntries();
    *str_ << entries.size() << endl;
    for(unsigned i = 0; i < entries.size(); i++) {
        *str_ << util_->showString(entries[i]->word) << endl;
        for(unsigned j = 0; j < entries[i]->prons.size(); j++) {
            if(j!=0) *str_ << " ";
            *str_ << util_->showString(entries[i]->prons[j]);
        }
        *str_ << endl << (entries[i]->inDict?'t':'f') << endl;
        writeModel(entries[i]->pronMod);
    }
}

Dictionary * TextModelIO::readDictionary() {
    Dictionary * dict = new Dictionary(util_);
    string line, buff;
    // get the states
    vector<DictionaryState*> & states = dict->getStates();
    getline(*str_, line);
    states.resize(util_->parseInt(line.c_str()));
    for(unsigned i = 0; i < states.size(); i++) {
        DictionaryState * state = new DictionaryState();
        getline(*str_, line);
        istringstream iss(line);
        iss >> buff;
        state->failure = util_->parseInt(buff.c_str());
        while(iss >> buff) {
            pair<KyteaChar,unsigned> p;
            p.first = util_->mapChar(buff.c_str());
            if(!(iss >> buff))
                throw runtime_error("Badly formed model (goto character without a destination)");
            p.second = util_->parseInt(buff.c_str());
            state->gotos.push_back(p);
        }
        sort(state->gotos.begin(), state->gotos.end());
        getline(*str_, line);
        istringstream iss2(line);
        while(iss2 >> buff)
            state->output.push_back(util_->parseInt(buff.c_str()));
        getline(*str_, line);
        if(line.length() != 1)
            throw runtime_error("Badly formed model (branch indicator not found)");
        state->isBranch = (line[0] == 'b');
        states[i] = state;
    }
    // get the entries
    vector<DictionaryEntry*> & entries = dict->getEntries();
    getline(*str_, line);
    entries.resize(util_->parseInt(line.c_str()));
    for(unsigned i = 0; i < entries.size(); i++) {
        getline(*str_, line);
        DictionaryEntry* entry = new DictionaryEntry(util_->mapString(line));
        getline(*str_, line);
        istringstream iss(line);
        while(iss >> buff)
            entry->prons.push_back(util_->mapString(buff.c_str()));
        getline(*str_, line);
        if(line.length() != 1 || (line[0] != 'f' && line[0] != 't'))
            throw runtime_error("Badly formed model (expecting 't' or 'f' after pronunciations)");
        entry->inDict = (line[0] == 't');
        entry->pronMod = readModel();
        entries[i] = entry;
    }
    return dict;
}


void BinaryModelIO::writeConfig(const KyteaConfig & config) {
    *str_ << "KyTea " << MODEL_IO_VERSION << " B " << config.getEncodingString() << endl;

    writeBinary(config.getDoWS());
    writeBinary(config.getDoPE());
    writeBinary(config.getCharWindow());
    writeBinary(config.getCharN());
    writeBinary(config.getTypeWindow());
    writeBinary(config.getTypeN());
    writeBinary(config.getDictionaryN());
    writeBinary(config.getBias());
    writeBinary(config.getEpsilon());
    writeBinary(config.getSolverType());
    
    // write the character map
    writeString(config.getStringUtil()->serialize());

}

void BinaryModelIO::readConfig(KyteaConfig & config) {
    
    string line;
    getline(*str_,line); // ignore the header

    config.setDoWS(readBinary<bool>() && config.getDoWS());
    config.setDoPE(readBinary<bool>() && config.getDoPE());
    config.setCharWindow(readBinary<char>());
    config.setCharN(readBinary<char>());
    config.setTypeWindow(readBinary<char>());
    config.setTypeN(readBinary<char>());
    config.setDictionaryN(readBinary<char>());
    config.setBias(readBinary<float>());
    config.setEpsilon(readBinary<float>());
    config.setSolverType(readBinary<int>());
    
    config.getStringUtil()->unserialize(readString());
    
}

void BinaryModelIO::writeModel(const KyteaModel * mod) {
    
    // print the number of features, zero for empty models
    if(mod == 0) { 
        writeBinary((int)0);
        return;
    }

    // print the feature names
    const KyteaModel::FeatVec & names = mod->getNames();
    writeBinary((int)names.size()-1);
    for(unsigned i = 1; i < names.size(); i++)
        writeString(names[i]);

    const model * model_ = mod->getModel();

    int i;
    int nr_feature=model_->nr_feature;
    int n;
    const parameter& param = model_->param;

    if(model_->bias>=0)
    	n=nr_feature+1;
    else
    	n=nr_feature;
    int w_size = n;

    int nr_w;
    if(model_->nr_class==2 && model_->param.solver_type != MCSVM_CS)
    	nr_w=1;
    else
    	nr_w=model_->nr_class;

    writeBinary(param.solver_type);
    writeBinary(model_->nr_class);
    for(i=0; i<model_->nr_class; i++)
        writeBinary(model_->label[i]);

    writeBinary(nr_feature);

    writeBinary((float)model_->bias);

    for(i=0; i<w_size; i++)
    {
    	int j;
    	for(j=0; j<nr_w; j++) {
            writeBinary((float)model_->w[i*nr_w+j]);
        }
    }

}

KyteaModel * BinaryModelIO::readModel() {

	int i = readBinary<int>();
    if(i == 0)
        return 0;

	int nr_feature;
	int n;
	int nr_class;
	model *model_ = (model*)malloc(sizeof(model));
    KyteaModel * mod = new KyteaModel();
    mod->setModel(model_);
	parameter& param = model_->param;

    for( ; i > 0; i--) {
        KyteaString myStr = readKyteaString();
        mod->mapFeat(myStr);
    }
    
    param.solver_type = readBinary<int>();
    nr_class = readBinary<int>();
    model_->nr_class = nr_class;
	model_->label = (int*)malloc(sizeof(int)*nr_class);
	for(i=0;i<nr_class;i++)
        model_->label[i] = readBinary<int>();

    nr_feature = readBinary<int>();

    model_->bias = readBinary<float>();

	model_->nr_feature=nr_feature;
	if(model_->bias>=0)
		n=nr_feature+1;
	else
		n=nr_feature;
	int w_size = n;
	int nr_w;
	if(nr_class==2 && param.solver_type != MCSVM_CS)
		nr_w = 1;
	else
		nr_w = nr_class;

	model_->w=(double*)malloc(sizeof(double)*w_size*nr_w);
	for(i=0; i<w_size; i++) 
		for(int j=0; j<nr_w; j++) 
            model_->w[i*nr_w+j] = readBinary<float>(); 
    
    // read models shouldn't add any additional features
    mod->setAddFeatures(false);

    return mod;

}

void BinaryModelIO::writeDictionary(const Dictionary * dict) {
    // write the states
    const vector<DictionaryState*> & states = dict->getStates();
    writeBinary(states.size());
    for(unsigned i = 0; i < states.size(); i++) {
        const DictionaryState * state = states[i];
        writeBinary(state->failure);
        writeBinary(state->gotos.size());
        for(unsigned j = 0; j < state->gotos.size(); j++) {
            writeBinary(state->gotos[j].first);
            writeBinary(state->gotos[j].second);
        }
        writeBinary(state->output.size());
        for(unsigned j = 0; j < state->output.size(); j++) 
            writeBinary(state->output[j]);
        writeBinary(state->isBranch);
    }
    // write the entries
    const vector<DictionaryEntry*> & entries = dict->getEntries();
    writeBinary(entries.size());
    for(unsigned i = 0; i < entries.size(); i++) {
        const DictionaryEntry * entry = entries[i];
        writeString(entry->word);
        writeBinary(entry->prons.size());
        for(unsigned j = 0; j < entry->prons.size(); j++)
            writeString(entry->prons[j]);
        writeBinary(entry->inDict);
        writeModel(entry->pronMod);
    }
}

Dictionary * BinaryModelIO::readDictionary() {
    Dictionary * dict = new Dictionary(util_);
    string line, buff;
    // get the states
    vector<DictionaryState*> & states = dict->getStates();
    states.resize(readBinary<size_t>());
    for(unsigned i = 0; i < states.size(); i++) {
        DictionaryState * state = new DictionaryState();
        state->failure = readBinary<unsigned>();
        state->gotos.resize(readBinary<size_t>());
        for(unsigned j = 0; j < state->gotos.size(); j++) {
            state->gotos[j].first = readBinary<KyteaChar>();
            state->gotos[j].second = readBinary<unsigned>();
        }
        state->output.resize(readBinary<size_t>());
        for(unsigned j = 0; j < state->output.size(); j++) 
            state->output[j] = readBinary<unsigned>();
        state->isBranch = readBinary<bool>();
        states[i] = state;
    }
    // get the entries
    vector<DictionaryEntry*> & entries = dict->getEntries();
    entries.resize(readBinary<size_t>());
    for(unsigned i = 0; i < entries.size(); i++) {
        DictionaryEntry* entry = new DictionaryEntry(readKyteaString());
        entry->prons.resize(readBinary<size_t>());
        for(unsigned j = 0; j < entry->prons.size(); j++)
            entry->prons[j] = readKyteaString();
        entry->inDict = readBinary<bool>();
        entry->pronMod = readModel();
        entries[i] = entry;
    }
    return dict;
}
